@@ -24,6 +24,8 @@ OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE.
 """
 import os
+import random
+
 import cv2
 import numpy as np
 
@@ -92,6 +94,8 @@ class Dataset:
                     # line: "<image_path> xmin,ymin,xmax,ymax,class_id ..."
                     bboxes = line.strip().split()
                     image_path = bboxes[0]
+                    if not os.path.exists(image_path):
+                        continue
                     xywhc_s = np.zeros((len(bboxes) - 1, 5))
                     for i, bbox in enumerate(bboxes[1:]):
                         # bbox = "xmin,ymin,xmax,ymax,class_id"
@@ -109,6 +113,8 @@ class Dataset:
                 for line in txt:
                     # line: "<image_path>"
                     image_path = line.strip()
+                    if not os.path.exists(image_path):
+                        continue
                     root, _ = os.path.splitext(image_path)
                     with open(root + ".txt") as fd2:
                         bboxes = fd2.readlines()
@@ -122,6 +128,10 @@ class Dataset:
                                 bbox[0],
                             )
                         _dataset.append([image_path, xywhc_s])
+
+        if len(_dataset) == 0:
+            raise FileNotFoundError("Failed to find images")
+
         return _dataset
 
     def bboxes_to_ground_truth(self, bboxes):
@@ -193,50 +203,65 @@ class Dataset:
 
         return ground_truth
 
-    def preprocess_dataset(self, dataset):
+    def load_image_then_resize(self, dataset, output_size=None):
         """
         @param dataset:
             yolo: [image_path, [[x, y, w, h, class_id], ...]]
             converted_coco: unit=> pixel
                 [image_path, [[x, y, w, h, class_id], ...]]
 
-        @return image / 255, ground_truth
+        @return image / 255, bboxes
         """
-        image = cv2.imread(dataset[0])
-        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
+        # pylint: disable=bare-except
+        try:
+            image = cv2.imread(dataset[0])
+            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
+        except:
+            return None
+
         if self.dataset_type == "converted_coco":
             height, width, _ = image.shape
             dataset[1] = dataset[1] / np.array(
                 [width, height, width, height, 1]
             )
 
+        if output_size is None:
+            output_size = self.input_size
+
         resized_image, resized_bboxes = media.resize_image(
-            image, self.input_size, dataset[1]
+            image, output_size, dataset[1]
         )
-
-        if self.data_augmentation:
-            # TODO
-            # BoF functions
-            pass
-
         resized_image = np.expand_dims(resized_image / 255.0, axis=0)
-        ground_truth = self.bboxes_to_ground_truth(resized_bboxes)
 
-        return resized_image, ground_truth
+        return resized_image, resized_bboxes
 
     def _next_data(self):
         for _ in range(5):
             _dataset = self.dataset[self.count]
-
             self.count += 1
             if self.count == len(self.dataset):
                 np.random.shuffle(self.dataset)
                 self.count = 0
 
-            if os.path.exists(_dataset[0]):
-                return _dataset
+            ret = self.load_image_then_resize(_dataset)
+            if ret is not None:
+                return ret
 
-        raise KeyError("Image search failed five times in a row")
+        raise FileNotFoundError("Failed to find images")
+
+    def _next_random_augmentation_data(self):
+        if random.random() < 0.2:
+            _prob = random.random()
+            if _prob < 0.25:
+                _dataset = cut_out(self._next_data())
+            elif _prob < 0.5:
+                _dataset = mix_up(self._next_data(), self._next_data())
+            else:
+                _dataset = mosaic(*[self._next_data() for _ in range(4)])
+        else:
+            _dataset = self._next_data()
+
+        return _dataset
 
     def __iter__(self):
         self.count = 0
@@ -252,16 +277,176 @@ class Dataset:
             batch_x = []
             _batch_y = [[] for _ in range(len(self.grid_size))]
             for _ in range(self.batch_size):
-                x, y = self.preprocess_dataset(self._next_data())
+                if self.data_augmentation:
+                    _dataset = self._next_random_augmentation_data()
+                else:
+                    _dataset = self._next_data()
+                x = _dataset[0]
+                y = self.bboxes_to_ground_truth(_dataset[1])
                 batch_x.append(x)
                 for i, _y in enumerate(y):
                     _batch_y[i].append(_y)
             batch_x = np.concatenate(batch_x, axis=0)
             batch_y = [np.concatenate(b_y, axis=0) for b_y in _batch_y]
         else:
-            batch_x, batch_y = self.preprocess_dataset(self._next_data())
+            if self.data_augmentation:
+                _dataset = self._next_random_augmentation_data()
+            else:
+                _dataset = self._next_data()
+            batch_x = _dataset[0]
+            batch_y = self.bboxes_to_ground_truth(_dataset[1])
 
+        # batch_x == Dim(batch, input_size, input_size, channels)
+        # batch_y[0] == Dim(batch, grid_size, grid_size, anchors, bboxes)
         return batch_x, batch_y
 
     def __len__(self):
         return len(self.dataset)
+
+
+def cut_out(dataset):
+    """
+    @parma dataset: image(float), bboxes
+            bboxes = [image_path, [[x, y, w, h, class_id], ...]]
+    """
+    _size = dataset[0].shape[1]
+    for bbox in dataset[1]:
+        if random.random() < 0.5:
+            _pixel_bbox = [int(pos * _size) for pos in bbox[0:4]]
+            _x_min = _pixel_bbox[0] - (_pixel_bbox[2] // 2)
+            _y_min = _pixel_bbox[1] - (_pixel_bbox[3] // 2)
+            _cut_out_width = _pixel_bbox[2] // 4
+            _cut_out_height = _pixel_bbox[3] // 4
+            _x_offset = (
+                int((_pixel_bbox[2] - _cut_out_width) * random.random())
+                + _x_min
+            )
+            _y_offset = (
+                int((_pixel_bbox[3] - _cut_out_height) * random.random())
+                + _y_min
+            )
+            dataset[0][
+                :,
+                _y_offset : _y_offset + _cut_out_height,
+                _x_offset : _x_offset + _cut_out_width,
+                :,
+            ] = 0.5
+
+    return dataset
+
+
+def mix_up(dataset0, dataset1, alpha=0.2):
+    return (
+        (dataset0[0] * alpha + dataset1[0] * (1 - alpha)),
+        (np.concatenate((dataset0[1], dataset1[1]), axis=0)),
+    )
+
+
+def mosaic(dataset0, dataset1, dataset2, dataset3):
+    size = dataset0[0].shape[1]
+    image = np.empty((1, size, size, 3))
+    bboxes = []
+
+    partition_x = int((random.random() * 0.6 + 0.2) * size)
+    partition_y = int((random.random() * 0.6 + 0.2) * size)
+
+    x_offset = [0, partition_x, 0, partition_x]
+    y_offset = [0, 0, partition_y, partition_y]
+
+    left = [
+        (size - partition_x) // 2,
+        partition_x // 2,
+        (size - partition_x) // 2,
+        partition_x // 2,
+    ]
+    right = [
+        left[0] + partition_x,
+        left[1] + size - partition_x,
+        left[2] + partition_x,
+        left[3] + size - partition_x,
+    ]
+    top = [
+        (size - partition_y) // 2,
+        (size - partition_y) // 2,
+        partition_y // 2,
+        partition_y // 2,
+    ]
+    down = [
+        top[0] + partition_y,
+        top[1] + partition_y,
+        top[2] + size - partition_y,
+        top[3] + size - partition_y,
+    ]
+
+    image[:, :partition_y, :partition_x, :] = dataset0[0][
+        :, top[0] : down[0], left[0] : right[0], :,
+    ]
+    image[:, :partition_y, partition_x:, :] = dataset1[0][
+        :, top[1] : down[1], left[1] : right[1], :,
+    ]
+    image[:, partition_y:, :partition_x, :] = dataset2[0][
+        :, top[2] : down[2], left[2] : right[2], :,
+    ]
+    image[:, partition_y:, partition_x:, :] = dataset3[0][
+        :, top[3] : down[3], left[3] : right[3], :,
+    ]
+
+    for i, _bboxes in enumerate(
+        (dataset0[1], dataset1[1], dataset2[1], dataset3[1])
+    ):
+        for bbox in _bboxes:
+            pixel_bbox = bbox[0:4] * size
+            x_min = int(pixel_bbox[0] - pixel_bbox[2] // 2)
+            y_min = int(pixel_bbox[1] - pixel_bbox[3] // 2)
+            x_max = int(pixel_bbox[0] + pixel_bbox[2] // 2)
+            y_max = int(pixel_bbox[1] + pixel_bbox[3] // 2)
+
+            class_id = bbox[4]
+
+            if x_min > right[i]:
+                continue
+            if y_min > down[i]:
+                continue
+            if x_max < left[i]:
+                continue
+            if y_max < top[i]:
+                continue
+
+            if x_max > right[i]:
+                x_max = right[i]
+            if y_max > down[i]:
+                y_max = down[i]
+            if x_min < left[i]:
+                x_min = left[i]
+            if y_min < top[i]:
+                y_min = top[i]
+
+            x_min -= left[i]
+            x_max -= left[i]
+            y_min -= top[i]
+            y_max -= top[i]
+
+            if x_min + 3 > x_max:
+                continue
+
+            if y_min + 3 > y_max:
+                continue
+
+            bboxes.append(
+                np.array(
+                    [
+                        [
+                            ((x_min + x_max) / 2 + x_offset[i]) / size,
+                            ((y_min + y_max) / 2 + y_offset[i]) / size,
+                            (x_max - x_min) / size,
+                            (y_max - y_min) / size,
+                            class_id,
+                        ],
+                    ]
+                )
+            )
+
+    if len(bboxes) == 0:
+        return dataset0
+
+    return image, np.concatenate(bboxes, axis=0)
